import logging

import numpy as np

from gunpowder.array import Array
from gunpowder.array_spec import ArraySpec
from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.ext import dvision
from gunpowder.profiling import Timing
from gunpowder.roi import Roi

from .batch_provider import BatchProvider

logger = logging.getLogger(__name__)


class DvidSource(BatchProvider):
    """A DVID array source.

    Provides arrays from DVID servers for each array key given.

    Args:

        hostname (``string``):

            The name of the DVID server.

        port (``int``):

            The port of the DVID server.

        uuid (``string``):

            The UUID of the DVID node to use.

        datasets (``dict``, :class:`ArrayKey` -> ``string``):

            Dictionary mapping array keys to DVID data instance names that this
            source offers.

        masks (``dict``, :class:`ArrayKey` -> ``string``, optional):

            Dictionary of array keys to DVID ROI instance names. This will
            create binary masks from DVID ROIs.

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            An optional dictionary of array keys to specs to overwrite the
            array specs automatically determined from the DVID server. This is
            useful to set ``voxel_size``, for example. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.
    """

    def __init__(self, hostname, port, uuid, datasets, masks=None, array_specs=None):
        self.hostname = hostname
        self.port = port
        self.url = "http://{}:{}".format(self.hostname, self.port)
        self.uuid = uuid

        self.datasets = datasets
        self.masks = masks if masks is not None else {}

        self.array_specs = array_specs if array_specs is not None else {}

        self.ndims = None

    def setup(self):
        for array_key, _ in self.datasets.items():
            spec = self.__get_spec(array_key)
            self.provides(array_key, spec)

        for array_key, _ in self.masks.items():
            spec = self.__get_mask_spec(array_key)
            self.provides(array_key, spec)

        logger.info("DvidSource.spec:\n%s", self.spec)

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        batch = Batch()

        for array_key, request_spec in request.array_specs.items():
            logger.debug("Reading %s in %s...", array_key, request_spec.roi)

            voxel_size = self.spec[array_key].voxel_size

            # scale request roi to voxel units
            dataset_roi = request_spec.roi / voxel_size

            # shift request roi into dataset
            dataset_roi = dataset_roi - self.spec[array_key].roi.offset / voxel_size

            # create array spec
            array_spec = self.spec[array_key].copy()
            array_spec.roi = request_spec.roi

            # read the data
            if array_key in self.datasets:
                data = self.__read_array(self.datasets[array_key], dataset_roi)
            elif array_key in self.masks:
                data = self.__read_mask(self.masks[array_key], dataset_roi)
            else:
                assert False, (
                    "Encountered a request for %s that is neither a volume "
                    "nor a mask." % array_key
                )

            # add array to batch
            batch.arrays[array_key] = Array(data, array_spec)

        logger.debug("done")

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def __get_info(self, array_key):
        if array_key in self.datasets:
            data = dvision.DVIDDataInstance(
                self.hostname, self.port, self.uuid, self.datasets[array_key]
            )

        elif array_key in self.masks:
            data = dvision.DVIDRegionOfInterest(
                self.hostname, self.port, self.uuid, self.masks[array_key]
            )

        else:
            assert False, "Encountered a request that is neither a volume nor a mask."

        return data.info

    def __get_spec(self, array_key):
        info = self.__get_info(array_key)

        roi_min = info["Extended"]["MinPoint"]
        if roi_min is not None:
            roi_min = Coordinate(roi_min[::-1])
        roi_max = info["Extended"]["MaxPoint"]
        if roi_max is not None:
            roi_max = Coordinate(roi_max[::-1])

        data_roi = Roi(offset=roi_min, shape=(roi_max - roi_min))
        data_dims = Coordinate(data_roi.shape)

        if self.ndims is None:
            self.ndims = len(data_dims)
        else:
            assert self.ndims == len(data_dims)

        if array_key in self.array_specs:
            spec = self.array_specs[array_key].copy()
        else:
            spec = ArraySpec()

        if spec.voxel_size is None:
            spec.voxel_size = Coordinate(info["Extended"]["VoxelSize"])

        if spec.roi is None:
            spec.roi = data_roi * spec.voxel_size

        data_dtype = dvision.DVIDDataInstance(
            self.hostname, self.port, self.uuid, self.datasets[array_key]
        ).dtype

        if spec.dtype is not None:
            assert spec.dtype == data_dtype, (
                "dtype %s provided in array_specs for %s, "
                "but differs from instance %s dtype %s"
                % (
                    self.array_specs[array_key].dtype,
                    array_key,
                    self.datasets[array_key],
                    data_dtype,
                )
            )
        else:
            spec.dtype = data_dtype

        if spec.interpolatable is None:
            spec.interpolatable = np.issubdtype(spec.dtype, np.floating) or (
                spec.dtype == np.uint8
            )
            logger.warning(
                "WARNING: You didn't set 'interpolatable' for %s. "
                "Based on the dtype %s, it has been set to %s. "
                "This might not be what you want.",
                array_key,
                spec.dtype,
                spec.interpolatable,
            )

        return spec

    def __get_mask_spec(self, mask_key):
        # create initial array spec

        if mask_key in self.array_specs:
            spec = self.array_specs[mask_key].copy()
        else:
            spec = ArraySpec()

        # get voxel size

        if spec.voxel_size is None:
            voxel_size = None
            for array_key in self.datasets:
                if voxel_size is None:
                    voxel_size = self.spec[array_key].voxel_size
                else:
                    assert voxel_size == self.spec[array_key].voxel_size, (
                        "No voxel size was given for mask %s, and the voxel "
                        "sizes of the volumes %s are not all the same. I don't "
                        "know what voxel size to use to create the mask."
                        % (mask_key, self.datasets.keys())
                    )

            spec.voxel_size = voxel_size

        # get ROI

        if spec.roi is None:
            for array_key in self.datasets:
                roi = self.spec[array_key].roi

                if spec.roi is None:
                    spec.roi = roi.copy()
                else:
                    spec.roi = roi.union(spec.roi)

        # set interpolatable

        if spec.interpolatable is None:
            spec.interpolatable = False

        # set datatype

        if spec.dtype is not None and spec.dtype != np.uint8:
            logger.warn(
                "Ignoring dtype in array_spec for %s, only np.uint8 "
                "is allowed for masks.",
                mask_key,
            )
        spec.dtype = np.uint8

        return spec

    def __read_array(self, instance, roi):
        data_instance = dvision.DVIDDataInstance(
            self.hostname, self.port, self.uuid, instance
        )

        return data_instance[roi.get_bounding_box()]

    def __read_mask(self, instance, roi):
        dvid_roi = dvision.DVIDRegionOfInterest(
            self.hostname, self.port, self.uuid, instance
        )

        return dvid_roi[roi.get_bounding_box()]

    def __repr__(self):
        return "DvidSource(hostname={}, port={}, uuid={}".format(
            self.hostname, self.port, self.uuid
        )



import copy
import logging
import random
import traceback
from typing import Optional

import numpy as np

from gunpowder.array import ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.coordinate import Coordinate
from gunpowder.graph import GraphKey
from gunpowder.graph_spec import GraphSpec
from gunpowder.provider_spec import ProviderSpec

logger = logging.getLogger(__name__)


class BatchRequestError(Exception):
    def __init__(
        self, provider, request, batch, original_traceback: Optional[list[str]] = None
    ):
        self.provider = provider
        self.request = request
        self.batch = batch
        self.original_traceback = original_traceback

    def __str__(self):
        return (
            f"Exception in {self.provider.name()} while processing request"
            f"{self.request}"
            "Batch returned so far:\n"
            f"{self.batch}" + ("\n\n" + "".join(self.original_traceback))
            if self.original_traceback is not None
            else ""
        )


class BatchProvider(object):
    """Superclass for all nodes in a `gunpowder` graph.

    A :class:`BatchProvider` provides :class:`Batches<Batch>` containing
    :class:`Arrays<Array>` and/or :class:`Graph`. The available data is
    specified in a :class:`ProviderSpec` instance, accessible via :attr:`spec`.

    To create a new node, subclass this class and implement (at least)
    :func:`setup` and :func:`provide`.

    A :class:`BatchProvider` can be linked to any number of other
    :class:`BatchProviders<BatchProvider>` upstream. If your node accepts
    exactly one upstream provider, consider subclassing :class:`BatchFilter`
    instead.
    """

    def add_upstream_provider(self, provider):
        self.get_upstream_providers().append(provider)
        return provider

    def remove_upstream_providers(self):
        self.upstream_providers = []

    def get_upstream_providers(self):
        if not hasattr(self, "upstream_providers"):
            self.upstream_providers = []
        return self.upstream_providers

    @property
    def remove_placeholders(self):
        if not hasattr(self, "_remove_placeholders"):
            return True
        return self._remove_placeholders

    def setup(self):
        """To be implemented in subclasses.

        Called during initialization of the DAG. Callees can assume that all
        upstream providers are set up already.

        In setup, call :func:`provides` to announce the arrays and points
        provided by this node.
        """
        raise NotImplementedError("Class %s does not implement 'setup'" % self.name())

    def teardown(self):
        """To be implemented in subclasses.

        Called during destruction of the DAG. Subclasses should use this to
        stop worker processes, if they used some.
        """
        pass

    def provides(self, key, spec):
        """Introduce a new output provided by this :class:`BatchProvider`.

        Implementations should call this in their :func:`setup` method, which
        will be called when the pipeline is build.

        Args:

            key (:class:`ArrayKey` or :class:`GraphKey`):

                The array or point set key provided.

            spec (:class:`ArraySpec` or :class:`GraphSpec`):

                The spec of the array or point set provided.
        """

        logger.debug("Current spec of %s:\n%s", self.name(), self.spec)

        if self.spec is None:
            self._spec = ProviderSpec()

        assert (
            key not in self.spec
        ), "Node %s is trying to add spec for %s, but is already " "provided." % (
            type(self).__name__,
            key,
        )

        self.spec[key] = copy.deepcopy(spec)
        self.provided_items.append(key)

        logger.debug("%s provides %s with spec %s", self.name(), key, spec)

    def _init_spec(self):
        if not hasattr(self, "_spec"):
            self._spec = None

    def internal_teardown(self):
        logger.debug("Resetting spec of %s", self.name())
        self._spec = None
        self._provided_items = []

        self.teardown()

    @property
    def spec(self):
        """Get the :class:`ProviderSpec` of this :class:`BatchProvider`.

        Note that the spec is only available after the pipeline has been build.
        Before that, it is ``None``.
        """
        self._init_spec()
        return self._spec

    @property
    def provided_items(self):
        """Get a list of the keys provided by this :class:`BatchProvider`.

        This list is only available after the pipeline has been build. Before
        that, it is empty.
        """

        if not hasattr(self, "_provided_items"):
            self._provided_items = []

        return self._provided_items

    def remove_provided(self, request):
        """Remove keys from `request` that are provided by this
        :class:`BatchProvider`.
        """

        for key in self.provided_items:
            if key in request:
                del request[key]

    def request_batch(self, request):
        """Request a batch from this provider.

        Args:

            request (:class:`BatchRequest`):

                A request containing (possibly partial)
                :class:`ArraySpecs<ArraySpec>` and
                :class:`GraphSpecs<GraphSpec>`.
        """

        batch = None

        try:
            self.set_seeds(request)

            logger.debug("%s got request %s", self.name(), request)

            self.check_request_consistency(request)

            upstream_request = request.copy()
            if self.remove_placeholders:
                upstream_request.remove_placeholders()
            batch = self.provide(upstream_request)

            request.remove_placeholders()

            self.check_batch_consistency(batch, request)

            self.remove_unneeded(batch, request)

            logger.debug("%s provides %s", self.name(), batch)

        except Exception as e:
            tb = traceback.format_exception(type(e), e, e.__traceback__)
            if isinstance(e, BatchRequestError):
                tb = tb[-1:]
            raise BatchRequestError(
                self, request, batch, original_traceback=tb
            ) from None

        return batch

    def set_seeds(self, request):
        seed = request.random_seed
        random.seed(seed)
        # augment uses numpy for its randomness
        np.random.seed(seed)

    def check_request_consistency(self, request):
        for key, request_spec in request.items():
            assert (
                key in self.spec
            ), "%s: Asked for %s which this node does not provide" % (self.name(), key)
            assert (
                isinstance(request_spec, ArraySpec)
                or isinstance(request_spec, GraphSpec)
                or isinstance(request_spec, GraphSpec)
            ), "spec for %s is of type%s" % (key, type(request_spec))

            provided_spec = self.spec[key]

            provided_roi = provided_spec.roi
            request_roi = request_spec.roi

            if provided_roi is not None:
                assert provided_roi.contains(
                    request_roi
                ), "%s: %s's ROI %s outside of my ROI %s" % (
                    self.name(),
                    key,
                    request_roi,
                    provided_roi,
                )

            if isinstance(key, ArrayKey):
                if request_spec.voxel_size is not None:
                    assert provided_spec.voxel_size == request_spec.voxel_size, (
                        "%s: voxel size %s requested for %s, but this node provides %s"
                        % (
                            self.name(),
                            request_spec.voxel_size,
                            key,
                            provided_spec.voxel_size,
                        )
                    )

                if request_roi is not None and provided_spec.voxel_size is not None:
                    for d in range(request_roi.dims):
                        assert (
                            request_roi.shape[d] % provided_spec.voxel_size[d] == 0
                        ), (
                            "in request %s, dimension %d of request %s is not a multiple of voxel_size %d"
                            % (request, d, key, provided_spec.voxel_size[d])
                        )

            if isinstance(key, GraphKey):
                if request_spec.directed is not None:
                    assert request_spec.directed == provided_spec.directed, (
                        f"asked for {key}:  directed={request_spec.directed} but "
                        f"{self.name()} provides directed={provided_spec.directed}"
                    )

    def check_batch_consistency(self, batch, request):
        for array_key, request_spec in request.array_specs.items():
            assert (
                array_key in batch.arrays
            ), "%s requested, but %s did not provide it." % (array_key, self.name())
            array = batch.arrays[array_key]
            assert (
                array.spec.roi == request_spec.roi
            ), "%s ROI %s requested, but ROI %s provided by %s." % (
                array_key,
                request_spec.roi,
                array.spec.roi,
                self.name(),
            )
            assert (
                array.spec.voxel_size == self.spec[array_key].voxel_size
            ), "voxel size of %s announced, but %s " "delivered for %s" % (
                self.spec[array_key].voxel_size,
                array.spec.voxel_size,
                array_key,
            )
            # ensure that the spatial dimensions are the same (other dimensions
            # on top are okay, e.g., for affinities)
            if request_spec.roi is not None:
                dims = request_spec.roi.dims
                data_shape = Coordinate(array.data.shape[-dims:])
                voxel_size = self.spec[array_key].voxel_size
                assert data_shape == request_spec.roi.shape / voxel_size, (
                    "%s ROI %s requested, but size of array is %s*%s=%s provided by %s."
                    % (
                        array_key,
                        request_spec.roi,
                        data_shape,
                        voxel_size,
                        data_shape * voxel_size,
                        self.name(),
                    )
                )
            if request_spec.dtype is not None:
                assert (
                    batch[array_key].data.dtype == request_spec.dtype
                ), "dtype of array %s (%s) does not match requested dtype %s by %s" % (
                    array_key,
                    batch[array_key].data.dtype,
                    request_spec.dtype,
                    self.name(),
                )

        for graph_key, request_spec in request.graph_specs.items():
            assert (
                graph_key in batch.graphs
            ), "%s requested, but %s did not provide it." % (graph_key, self.name())
            graph = batch.graphs[graph_key]
            assert (
                graph.spec.roi == request_spec.roi
            ), "%s ROI %s requested, but ROI %s provided by %s." % (
                graph_key,
                request_spec.roi,
                graph.spec.roi,
                self.name(),
            )

            if request_spec.directed is not None:
                assert request_spec.directed == graph.directed, (
                    f"Recieved {graph_key}:  directed={graph.directed} but "
                    f"{self.name()} should provide directed={request_spec.directed}"
                )

            for node in graph.nodes:
                contained = graph.spec.roi.contains(node.location)
                dangling = not contained and all(
                    [graph.spec.roi.contains(v.location) for v in graph.neighbors(node)]
                )
                assert contained or dangling, (
                    f"graph {graph_key} provided by {self.name()} with ROI {graph.spec.roi} "
                    f"contain point at {node.location} which is neither contained nor "
                    f"'dangling'"
                )

    def remove_unneeded(self, batch, request):
        batch_keys = set(list(batch.arrays.keys()) + list(batch.graphs.keys()))
        for key in batch_keys:
            if key not in request:
                del batch[key]

    def enable_placeholders(self):
        self._remove_placeholders = False

    def provide(self, request):
        """To be implemented in subclasses.

        This function takes a :class:`BatchRequest` and should return the
        corresponding :class:`Batch`.

        Args:

            request(:class:`BatchRequest`):

                The request to process.
        """
        raise NotImplementedError("Class %s does not implement 'provide'" % self.name())

    def name(self):
        return type(self).__name__

    def __repr__(self):
        return self.name() + ", providing: " + str(self.spec)

    def __add__(self, other):
        """Support ``self + other`` operator. Return a :class:`Pipeline`."""
        from gunpowder import Pipeline

        if isinstance(other, BatchProvider):
            other = Pipeline(other)

        if not isinstance(other, Pipeline):
            raise RuntimeError(
                f"Don't know how to add {type(other)} to BatchProvider {self.name()}"
            )

        return Pipeline(self) + other

    def __radd__(self, other):
        """Support ``other + self`` operator. Return a :class:`Pipeline`."""
        from gunpowder import Pipeline

        if isinstance(other, BatchProvider):
            return Pipeline(other) + Pipeline(self)

        if isinstance(other, tuple):
            return other + Pipeline(self)

        raise RuntimeError(
            f"Don't know how to radd {type(other)} to BatchProvider{self.name()}"
        )



import os

from gunpowder.ext import h5py

from .hdf5like_write_base import Hdf5LikeWrite


class Hdf5Write(Hdf5LikeWrite):
    """Assemble arrays of passing batches in one HDF5 file. This is useful to
    store chunks produced by :class:`Scan` on disk without keeping the larger
    array in memory. The ROIs of the passing arrays will be used to determine
    the position where to store the data in the dataset.

    Args:

        dataset_names (``dict``, :class:`ArrayKey` -> ``string``):

            A dictionary from array keys to names of the datasets to store them
            in.

        output_dir (``string``):

            The directory to save the HDF5 file. Will be created, if it does
            not exist.

        output_filename (``string``):

            The output filename of the container. Will be created, if it does
            not exist, otherwise data is overwritten in the existing container.

        compression_type (``string`` or ``int``):

            Compression strategy.  Legal values are ``gzip``, ``szip``,
            ``lzf``. If an integer between 1 and 10, this indicates ``gzip``
            compression level.

        dataset_dtypes (``dict``, :class:`ArrayKey` -> data type):

            A dictionary from array keys to datatype (eg. ``np.int8``). If
            given, arrays are stored using this type. The original arrays
            within the pipeline remain unchanged.
    """

    def _open_file(self, filename):
        if os.path.exists(filename):
            return h5py.File(filename, "r+")
        else:
            return h5py.File(filename, "w")



import logging
import multiprocessing
import time

from gunpowder.batch import Batch
from gunpowder.ext import daisy
from gunpowder.nodes.batch_filter import BatchFilter
from gunpowder.roi import Roi

logger = logging.getLogger(__name__)


class DaisyRequestBlocks(BatchFilter):
    """Iteratively requests batches similar to ``reference`` from upstream
    providers, with their ROIs set to blocks distributed by ``daisy``.

    The ROIs of the array or point specs in the reference can be set to either
    the block's ``read_roi`` or ``write_roi``, see parameter ``roi_map``.

    The batch request to this node has to be empty, as there is no guarantee
    that this node will get to process all chunks required to fulfill a
    particular batch request.

    Args:

        reference (:class:`BatchRequest`):

            A reference :class:`BatchRequest`. This request will be shifted
            according to blocks distributed by ``daisy``.

        roi_map (``dict`` from :class:`ArrayKey` or :class:`GraphKey` to
        ``string``):

            A map indicating which daisy block ROI (``read_roi`` or
            ``write_roi``) to use for which item in the reference request.

        num_workers (``int``, optional):

            If set to >1, upstream requests are made in parallel with that
            number of workers.

        block_done_callback (function, optional):

            If given, will be called with arguments ``(block, start,
            duration)`` for each block that was processed. ``start`` and
            ``duration`` will be given in seconds, as in ``start =
            time.time()`` and ``duration = time.time() - start``, right before
            and after a block gets processed.

            This callback can be used to log blocks that have successfully
            finished processing, which can be used in ``check_function`` of
            ``daisy.run_blockwise`` to skip already processed blocks in
            repeated runs.
    """

    def __init__(self, reference, roi_map, num_workers=1, block_done_callback=None):
        self.reference = reference
        self.roi_map = roi_map
        self.num_workers = num_workers
        self.block_done_callback = block_done_callback

        if num_workers > 1:
            self.request_queue = multiprocessing.Queue(maxsize=0)

    def provide(self, request):
        empty_request = len(request) == 0
        if not empty_request:
            raise RuntimeError("requests made to DaisyRequestBlocks have to be empty")

        if self.num_workers > 1:
            self.workers = [
                multiprocessing.Process(target=self.__get_chunks)
                for _ in range(self.num_workers)
            ]

            for worker in self.workers:
                worker.start()

            for worker in self.workers:
                worker.join()

        else:
            self.__get_chunks()

        return Batch()

    def __get_chunks(self):
        daisy_client = daisy.Client()

        while True:
            with daisy_client.acquire_block() as block:
                if block is None:
                    return

                logger.info("Processing block %s", block)
                start = time.time()

                chunk_request = self.reference.copy()

                for key, reference_spec in self.reference.items():
                    roi_type = self.roi_map.get(key, None)

                    if roi_type is None:
                        raise RuntimeError(
                            "roi_map does not map item %s to either 'read_roi' "
                            "or 'write_roi'" % key
                        )

                    if roi_type == "read_roi":
                        chunk_request[key].roi = Roi(
                            block.read_roi.offset, block.read_roi.shape
                        )
                    elif roi_type == "write_roi":
                        chunk_request[key].roi = Roi(
                            block.write_roi.offset, block.write_roi.shape
                        )
                    else:
                        raise RuntimeError(
                            "%s is not a vaid ROI type (read_roi or write_roi)"
                        )

                self.get_upstream_provider().request_batch(chunk_request)

                end = time.time()
                if self.block_done_callback:
                    self.block_done_callback(block, start, end - start)



import logging
from typing import Any, Dict, Optional, Union

from gunpowder.array import Array, ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.ext import torch
from gunpowder.nodes.generic_predict import GenericPredict

logger = logging.getLogger(__name__)


class Predict(GenericPredict):
    """Torch implementation of :class:`gunpowder.nodes.Predict`.

    Args:

        model (subclass of ``torch.nn.Module``):

            The model to use for prediction.

        inputs (``dict``, ``string`` or ``int`` -> :class:`ArrayKey`):

            Dictionary from the position (for args) and names (for kwargs) of input
            tensors (argument names of the ``forward`` method) in the model to array keys.

        outputs (``dict``, ``string`` or ``int`` -> :class:`ArrayKey`):

            Dictionary from the names of tensors in the network to array
            keys. If the key is a string, the tensor will be retrieved
            by checking the model for an attribute with the key as its name.
            If the key is an integer, it is interpreted as a tuple index of
            the outputs of the network.
            New arrays will be generated by this node for each entry (if
            requested downstream).

        array_specs (``dict``, :class:`ArrayKey` -> :class:`ArraySpec`, optional):

            Used to set the specs of generated arrays (``outputs``). This is
            useful to set the ``voxel_size``, for example, if they differ from
            the voxel size of the input arrays. Only fields that are not
            ``None`` in the given :class:`ArraySpec` will be used.

        checkpoint: (``string``, optional):

            An optional path to the saved parameters for your torch module.
            These will be loaded and used for prediction if provided.

        device (``string``, optional):

            Which device to use for prediction (``"cpu"`` or ``"cuda"``).
            Default is ``"cuda"``, which falls back to CPU if CUDA is not
            available.

        spawn_subprocess (bool, optional): Whether to run ``predict`` in a
            separate process. Default is false.
    """

    def __init__(
        self,
        model,
        inputs: Dict[Union[str, int], ArrayKey],
        outputs: Dict[Union[str, int], ArrayKey],
        array_specs: Optional[Dict[ArrayKey, ArraySpec]] = None,
        checkpoint: Optional[str] = None,
        device="cuda",
        spawn_subprocess=False,
    ):
        self.array_specs = array_specs if array_specs is not None else {}

        if model.training:
            logger.warning(
                "Model is in training mode during prediction. "
                "Consider using model.eval()"
            )

        super(Predict, self).__init__(
            inputs, outputs, array_specs, spawn_subprocess=spawn_subprocess
        )

        self.device_string = device
        self.device = None  # to be set in start()
        self.model = model
        self.checkpoint = checkpoint

        self.intermediate_layers: dict[ArrayKey, Any] = {}

    def start(self):
        # Issue #188
        self.use_cuda = torch.cuda.is_available() and self.device_string.__contains__(
            "cuda"
        )

        logger.info(f"Predicting on {'gpu' if self.use_cuda else 'cpu'}")
        self.device = torch.device(self.device_string if self.use_cuda else "cpu")
        try:
            self.model = self.model.to(self.device)
        except RuntimeError as e:
            raise RuntimeError(
                "Failed to move model to device. If you are using a child process "
                "to run your model, maybe you already initialized CUDA by sending "
                "your model to device in the main process."
            ) from e

        if self.checkpoint is not None:
            checkpoint = torch.load(self.checkpoint, map_location=self.device)
            if "model_state_dict" in checkpoint:
                self.model.load_state_dict(checkpoint["model_state_dict"])
            else:
                self.model.load_state_dict(checkpoint)

        self.register_hooks()

    def predict(self, batch, request):
        input_args, input_kwargs = self.get_inputs(batch)
        with torch.no_grad():
            out = self.model.forward(*input_args, **input_kwargs)
        outputs = self.get_outputs(out, request)
        self.update_batch(batch, request, outputs)

    def get_inputs(self, batch):
        model_args = [
            torch.as_tensor(batch[self.inputs[ii]].data, device=self.device)
            for ii in range(len(self.inputs))
            if ii in self.inputs
        ]
        model_kwargs = {
            key: torch.as_tensor(batch[value].data, device=self.device)
            for key, value in self.inputs.items()
            if isinstance(key, str)
        }
        return model_args, model_kwargs

    def register_hooks(self):
        for key in self.outputs:
            if isinstance(key, str):
                layer = getattr(self.model, key)
                layer.register_forward_hook(self.create_hook(key))

    def create_hook(self, key):
        def save_layer(module, input, output):
            self.intermediate_layers[key] = output

        return save_layer

    def get_outputs(self, module_out, request):
        outputs = {}
        if isinstance(module_out, tuple):
            module_outs = module_out
        else:
            module_outs = (module_out,)
        for key, value in self.outputs.items():
            if value in request:
                if isinstance(key, str):
                    outputs[value] = self.intermediate_layers[key]
                elif isinstance(key, int):
                    outputs[value] = module_outs[key]
        return outputs

    def update_batch(self, batch, request, requested_outputs):
        for array_key, tensor in requested_outputs.items():
            spec = self.spec[array_key].copy()
            spec.roi = request[array_key].roi
            batch.arrays[array_key] = Array(tensor.cpu().detach().numpy(), spec)

    def stop(self):
        pass



from __future__ import division, print_function

import logging
import random

import numpy as np

from gunpowder.coordinate import Coordinate
from gunpowder.roi import Roi

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class ShiftAugment(BatchFilter):
    def __init__(self, prob_slip=0, prob_shift=0, sigma=0, shift_axis=0, p=1.0):
        self.prob_slip = prob_slip
        self.prob_shift = prob_shift
        self.sigma = sigma
        self.shift_axis = shift_axis
        self.p = p

        self.ndim = None
        self.shift_sigmas = None
        self.shift_array = None
        self.lcm_voxel_size = None

    def skip_node(self, request):
        return random.random() > self.p

    def prepare(self, request):
        self.ndim = request.get_total_roi().dims
        assert self.shift_axis in range(self.ndim)

        try:
            self.shift_sigmas = tuple(self.sigma)
        except TypeError:
            self.shift_sigmas = [float(self.sigma)] * self.ndim
            self.shift_sigmas[self.shift_axis] = 0.0
            self.shift_sigmas = tuple(self.shift_sigmas)

        assert len(self.shift_sigmas) == self.ndim
        assert self.shift_sigmas[self.shift_axis] == 0.0

        has_nonzero = False
        for sigma in self.shift_sigmas:
            if sigma != 0.0:
                has_nonzero = True
                break
        assert has_nonzero

        if not request.array_specs:
            raise ValueError(
                "Request passed to Jitter node must contain at least one array key. "
                + "Check to make sure that Jitter node is not upstream of a RandomLocation node "
                + "with an ensure_nonempty argument."
            )

        self.lcm_voxel_size = self.spec.get_lcm_voxel_size(
            array_keys=request.array_specs.keys()
        )
        assert self.lcm_voxel_size

        roi_shape = request.get_total_roi().shape
        assert (
            roi_shape // self.lcm_voxel_size * self.lcm_voxel_size == roi_shape
        ), "total roi shape {} must be divisible by least common voxel size {}".format(
            roi_shape, self.lcm_voxel_size
        )
        roi_shape_adjusted = roi_shape // self.lcm_voxel_size
        shift_axis_len = roi_shape_adjusted[self.shift_axis]

        self.shift_array = self.construct_global_shift_array(
            shift_axis_len,
            self.shift_sigmas,
            self.prob_slip,
            self.prob_shift,
            self.lcm_voxel_size,
        )

        for key, spec in request.items():
            sub_shift_array = self.get_sub_shift_array(
                request.get_total_roi(),
                spec.roi,
                self.shift_array,
                self.shift_axis,
                self.lcm_voxel_size,
            )
            updated_roi = self.compute_upstream_roi(spec.roi, sub_shift_array)
            spec.roi.offset = updated_roi.offset
            spec.roi.shape = updated_roi.shape
            request[key] = spec

        deps = request
        return deps

    def process(self, batch, request):
        for array_key, array in batch.arrays.items():
            sub_shift_array = self.get_sub_shift_array(
                request.get_total_roi(),
                array.spec.roi,
                self.shift_array,
                self.shift_axis,
                self.lcm_voxel_size,
            )
            array.data = self.shift_and_crop(
                array.data,
                request[array_key].roi.shape,
                sub_shift_array,
                array.spec.voxel_size,
            )
            array.spec.roi = request[array_key].roi
            assert (
                request[array_key].roi.shape
                == Coordinate(array.data.shape) * self.lcm_voxel_size
            ), "request roi shape {} is not the same as generated array shape {}".format(
                request[array_key].roi.shape, array.data.shape
            )
            batch[array_key] = array

        for points_key, points in batch.graphs.items():
            sub_shift_array = self.get_sub_shift_array(
                request.get_total_roi(),
                points.spec.roi,
                self.shift_array,
                self.shift_axis,
                self.lcm_voxel_size,
            )
            points = self.shift_points(
                points,
                request[points_key].roi,
                sub_shift_array,
                self.shift_axis,
                self.lcm_voxel_size,
            )
            batch[points_key] = points

    def shift_and_crop(self, arr, roi_shape, sub_shift_array, voxel_size):
        """Shift an array received from upstream and crop it to the target downstream region

        :param arr: an array of upstream data to be shifted and cropped
        :param roi_shape: the shape of the downstream ROI
        :param sub_shift_array: the cropped section of the global shift array that applies to this specific request
        :param voxel_size: the voxel sizes of the data in the array
        :return: an array of shape roi_shape that contains the array to be passed downstream
        """

        array_shift_axis_len = arr.shape[self.shift_axis]
        sub_shift_array_len = len(sub_shift_array)
        assert (
            array_shift_axis_len % sub_shift_array_len == 0
        ), "array shift axis length {} is not divisible by the sub_shift_array length {}".format(
            arr.shape[self.shift_axis], sub_shift_array.shape[0]
        )

        voxel_ratio = array_shift_axis_len // sub_shift_array_len

        # assumption: each sub shift array element divides evenly by the voxel size
        rescaled_sub_shift_array = sub_shift_array // np.array(voxel_size, dtype=int)

        max_shift = rescaled_sub_shift_array.max(axis=0)
        batch = arr.copy()
        batch_view = np.moveaxis(batch, self.shift_axis, 0)
        for index, plane in enumerate(batch_view):
            adjusted_index = index // voxel_ratio
            shift = rescaled_sub_shift_array[adjusted_index, :] - max_shift
            shift = np.delete(shift, self.shift_axis, axis=0)
            assert len(shift) == plane.ndim
            plane = np.roll(plane, shift, axis=tuple(range(len(shift))))
            batch_view[index] = plane

        adjusted_roi_shape = Coordinate(roi_shape) // Coordinate(voxel_size)

        sl = tuple(slice(0, adjusted_roi_shape[index]) for index in range(self.ndim))
        return batch[sl]

    @staticmethod
    def shift_points(points, request_roi, sub_shift_array, shift_axis, lcm_voxel_size):
        """Shift a set of points received from upstream and crop out those not the the target downstream region

        :param points: the points from upstream
        :param request_roi: the downstream ROI
        :param sub_shift_array: the cropped section of the global shift array that applies to this specific request
        :param shift_axis: the axis to perform the shift along
        :param lcm_voxel_size: the least common voxel size for the arrays in the request
        :return a Graph object with the updated point locations and ROI
        """

        nodes = list(points.nodes)
        spec = points.spec
        shift_axis_start_pos = spec.roi.offset[shift_axis]

        for node in nodes:
            loc = node.location
            shift_axis_position = loc[shift_axis]
            shift_array_index = int(
                (shift_axis_position - shift_axis_start_pos)
                // lcm_voxel_size[shift_axis]
            )
            assert shift_array_index >= 0
            shift = Coordinate(sub_shift_array[shift_array_index])
            loc += shift
            if not request_roi.contains(loc):
                points.remove_node(node)

        points.spec.roi = request_roi
        return points

    @staticmethod
    def get_sub_shift_array(
        total_roi, item_roi, shift_array, shift_axis, lcm_voxel_size
    ):
        """Slices the global shift array to return the sub-shift array to shift an item in the request

        :param total_roi: the total roi of the request
        :param item_roi: the roi of the item (array or points) being shifted
        :param shift_array: the shift array for the total_roi
        :param shift_axis: the axis along which we are shifting
        :param lcm_voxel_size: the least common voxel size for the arrays in the request
        :return: the portion of the global shift array that should be used to shift the item
        """
        item_offset_from_total = item_roi.offset - total_roi.offset
        offset_in_shift_axis = (
            item_offset_from_total[shift_axis] // lcm_voxel_size[shift_axis]
        )
        len_in_shift_axis = item_roi.shape[shift_axis] // lcm_voxel_size[shift_axis]
        return shift_array[
            offset_in_shift_axis : offset_in_shift_axis + len_in_shift_axis
        ]

    @staticmethod
    def construct_global_shift_array(
        shift_axis_len, shift_sigmas, prob_slip, prob_shift, lcm_voxel_size
    ):
        """Sets the attribute variable self.shift_array

        :param shift_axis_len: the length of the shift axis
        :param shift_sigmas: the sigma to generate the normal distribution of shift amounts in each direction
        :param prob_slip: the probability of the slice shifting independently of all other slices
        :param prob_shift: the probability of the slice and all following slices shifting
        :param lcm_voxel_size: the least common voxel size of all the arrays in the request
        :return: the shift_array for the total_roi
        """
        # each row is one slice along shift axis
        shift_array = np.zeros(shape=(shift_axis_len, len(shift_sigmas)), dtype=int)
        base_shift = np.zeros(shape=len(shift_sigmas), dtype=int)
        assert prob_slip + prob_shift <= 1

        for shift_axis_position in range(shift_axis_len):
            r = random.random()
            slip = np.array(
                [
                    np.random.normal(scale=sigma / lcm_voxel_size[dimension])
                    for dimension, sigma in enumerate(shift_sigmas)
                ]
            )
            slip = np.rint(slip).astype(int)
            slip = slip * np.array(lcm_voxel_size, dtype=int)

            if r <= prob_slip:
                shift_array[shift_axis_position] = base_shift + slip
            elif r <= prob_slip + prob_shift:
                base_shift += slip
                shift_array[shift_axis_position] = base_shift
            else:
                shift_array[shift_axis_position] = base_shift

        return shift_array

    @staticmethod
    def compute_upstream_roi(request_roi, sub_shift_array):
        """Compute the ROI to pass upstream for a specific item (array or points) in a request

        :param request_roi: the downstream ROI passed to the Jitter node
        :param sub_shift_array: the portion of the global shift array that should be used to shift the item
        :return: the expanded ROI to pass upstream
        """

        max_shift = Coordinate(sub_shift_array.max(axis=0))
        min_shift = Coordinate(sub_shift_array.min(axis=0))

        downstream_offset = request_roi.offset
        upstream_offset = downstream_offset - max_shift
        upstream_shape = request_roi.shape + max_shift - min_shift
        return Roi(offset=upstream_offset, shape=upstream_shape)



import copy
import logging

import numpy as np

from gunpowder.array import Array
from gunpowder.morphology import enlarge_binary_map
from gunpowder.nodes.batch_filter import BatchFilter
from gunpowder.nodes.rasterize_graph import RasterizationSettings

logger = logging.getLogger(__name__)


class AddGtMaskExclusiveZone(BatchFilter):
    """Create ExclusizeZone mask for a binary map in batch and add it as
    array to batch.

    An ExclusiveZone mask is a bianry mask [0,1] where locations which lie
    within a given distance to the ON (=1) regions (surrounding the ON regions)
    of the given binary map are set to 0, whereas all the others are set to 1.

    Args:

        EZ_masks_to_binary_map(dict, :class:``ArrayKey``->:class:``ArrayKey``):
            Arrays of exclusive zones (keys of dict) to create for which
            binary mask (values of dict).

        gaussian_sigma_for_zone(float, optional): Defines extend of exclusive
            zone around ON region in binary map. Defaults to 1.

        rasterization_setting(:class:``RasterizationSettings``, optional): Which
            rasterization setting to use.
    """

    def __init__(
        self,
        EZ_masks_to_binary_map,
        gaussian_sigma_for_zone=1,
        rasterization_setting=None,
    ):
        self.EZ_masks_to_binary_map = EZ_masks_to_binary_map
        self.gaussian_sigma_for_zone = gaussian_sigma_for_zone
        if rasterization_setting is None:
            self.rasterization_setting = RasterizationSettings()
        else:
            self.rasterization_setting = rasterization_setting
        self.skip_next = False

    def setup(self):
        self.upstream_spec = self.get_upstream_provider().get_spec()
        self.spec = copy.deepcopy(self.upstream_spec)

        for EZ_mask_type, binary_map_type in self.EZ_masks_to_binary_map.items():
            if binary_map_type in self.upstream_spec.arrays:
                self.spec.arrays[EZ_mask_type] = self.spec.arrays[binary_map_type]

    def get_spec(self):
        return self.spec

    def prepare(self, request):
        self.EZ_masks_to_create = []
        for EZ_mask_type, binary_map_type in self.EZ_masks_to_binary_map.items():
            # do nothing if binary mask to create EZ mask around is not requested as well
            if EZ_mask_type in request.arrays:
                # assert that binary mask for which EZ mask is created for is requested
                assert (
                    binary_map_type in request.arrays
                ), "ExclusiveZone Mask for {}, can only be created if {} also requested.".format(
                    EZ_mask_type, binary_map_type
                )
                # assert that ROI of EZ lies within ROI of binary mask
                assert request.arrays[binary_map_type].contains(
                    request.arrays[EZ_mask_type]
                ), "EZ mask for {} requested for ROI outside of source's ({}) ROI.".format(
                    EZ_mask_type, binary_map_type
                )

                self.EZ_masks_to_create.append(EZ_mask_type)
                del request.arrays[EZ_mask_type]

        if len(self.EZ_masks_to_create) == 0:
            logger.warn("no ExclusiveZone Masks requested, will do nothing")
            self.skip_next = True

    def process(self, batch, request):
        # do nothing if no gt binary maps were requested
        if self.skip_next:
            self.skip_next = False
            return

        for EZ_mask_type in self.EZ_masks_to_create:
            binary_map_type = self.EZ_masks_to_binary_map[EZ_mask_type]
            binary_map = batch.arrays[binary_map_type].data
            resolution = batch.arrays[binary_map_type].resolution
            EZ_mask = self.__get_exclusivezone_mask(
                binary_map,
                shape_EZ_mask=request.arrays[EZ_mask_type].get_shape(),
                resolution=resolution,
            )

            batch.arrays[EZ_mask_type] = Array(
                data=EZ_mask, roi=request.arrays[EZ_mask_type], resolution=resolution
            )

    def __get_exclusivezone_mask(self, binary_map, shape_EZ_mask, resolution=None):
        """Exclusive zone surrounds every synapse. Created by enlarging the ON regions of given binary map
        with different gaussian filter, make it binary and subtract the original binary map from it
        """

        shape_diff = np.asarray(binary_map.shape - np.asarray(shape_EZ_mask))
        slices = [
            slice(diff, shape - diff)
            for diff, shape in zip(shape_diff, binary_map.shape)
        ]
        relevant_binary_map = binary_map[slices]

        BM_enlarged_binary = enlarge_binary_map(
            relevant_binary_map,
            marker_size_voxel=self.rasterization_setting.marker_size_voxel,
            voxel_size=resolution,
            marker_size_physical=self.rasterization_setting.marker_size_physical,
        )

        exclusive_zone = np.ones_like(BM_enlarged_binary) - (
            BM_enlarged_binary - relevant_binary_map
        )
        return exclusive_zone



import logging
import warnings
from collections.abc import MutableMapping
from typing import Union

from zarr import N5FSStore, N5Store
from zarr._storage.store import BaseStore

from gunpowder.array import ArrayKey
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.ext import ZarrFile
from gunpowder.roi import Roi

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class ZarrWrite(BatchFilter):
    """Assemble arrays of passing batches in one zarr container. This is useful
    to store chunks produced by :class:`Scan` on disk without keeping the
    larger array in memory. The ROIs of the passing arrays will be used to
    determine the position where to store the data in the dataset.

    Args:

        dataset_names (``dict``, :class:`ArrayKey` -> ``string``):

            A dictionary from array keys to names of the datasets to store them
            in.

        store (``string`` or ``BaseStore``):

            The directory to save the zarr container. Will be created, if it does
            not exist.

        compression_type (``string`` or ``int``):

            Compression strategy.  Legal values are ``gzip``, ``szip``,
            ``lzf``. If an integer between 1 and 10, this indicates ``gzip``
            compression level.

        dataset_dtypes (``dict``, :class:`ArrayKey` -> data type):

            A dictionary from array keys to datatype (eg. ``np.int8``). If
            given, arrays are stored using this type. The original arrays
            within the pipeline remain unchanged.
    """

    def __init__(
        self,
        dataset_names,
        output_dir=".",
        output_filename="output.hdf",
        compression_type=None,
        dataset_dtypes=None,
        store: Union[BaseStore, MutableMapping, str] = None,
    ):
        self.store = store if store is not None else f"{output_dir}/{output_filename}"
        if store is None:
            warnings.warn(
                "Argument 'output_dir' and `output_filename` will be replaced in v2.0, "
                "use 'store' instead",
                DeprecationWarning,
            )
        self.dataset_names = dataset_names
        self.compression_type = compression_type
        if dataset_dtypes is None:
            self.dataset_dtypes = {}
        else:
            self.dataset_dtypes = dataset_dtypes

        self.dataset_offsets: dict[ArrayKey, Coordinate] = {}

    def _get_voxel_size(self, dataset):
        if "resolution" not in dataset.attrs:
            return None
        if self._rev_metadata():
            return Coordinate(dataset.attrs["resolution"][::-1])
        else:
            return Coordinate(dataset.attrs["resolution"])

    def _get_offset(self, dataset):
        if "offset" not in dataset.attrs:
            return None
        if self._rev_metadata():
            return Coordinate(dataset.attrs["offset"][::-1])
        else:
            return Coordinate(dataset.attrs["offset"])

    def _set_voxel_size(self, dataset, voxel_size):
        if self._rev_metadata():
            dataset.attrs["resolution"] = voxel_size[::-1]
        else:
            dataset.attrs["resolution"] = voxel_size

    def _set_offset(self, dataset, offset):
        if self._rev_metadata():
            dataset.attrs["offset"] = offset[::-1]
        else:
            dataset.attrs["offset"] = offset

    def _rev_metadata(self):
        with ZarrFile(self.store, mode="a") as store:
            return isinstance(store.chunk_store, N5Store) or isinstance(
                store.chunk_store, N5FSStore
            )

    def _open_file(self, store):
        return ZarrFile(store, mode="a")

    def setup(self):
        for key in self.dataset_names.keys():
            self.updates(key, self.spec[key])
        self.enable_autoskip()

    def prepare(self, request):
        deps = BatchRequest()
        for key in self.dataset_names.keys():
            deps[key] = request[key]
        return deps

    def init_datasets(self, batch):
        with self._open_file(self.store) as data_file:
            for array_key, dataset_name in self.dataset_names.items():
                logger.debug("Initializing dataset for %s", array_key)

                assert array_key in self.spec, (
                    "Asked to store %s, but is not provided upstream." % array_key
                )
                assert array_key in batch.arrays, (
                    "Asked to store %s, but is not part of batch." % array_key
                )

                array = batch.arrays[array_key]
                dims = array.spec.roi.dims
                batch_shape = array.data.shape

                # if a dataset already exists, read its meta-information (if
                # present)
                if dataset_name in data_file:
                    offset = self._get_offset(data_file[dataset_name]) or Coordinate(
                        (0,) * dims
                    )

                else:
                    provided_roi = self.spec[array_key].roi

                    if provided_roi is None:
                        raise RuntimeError(
                            "Dataset %s does not exist in %s, and no ROI is "
                            "provided for %s. I don't know how to initialize "
                            "the dataset." % (dataset_name, self.store, array_key)
                        )

                    offset = provided_roi.offset
                    voxel_size = array.spec.voxel_size
                    data_shape = provided_roi.shape // voxel_size

                    logger.debug("Shape in voxels: %s", data_shape)
                    # add channel dimensions (if present)
                    data_shape = batch_shape[:-dims] + data_shape
                    logger.debug("Shape with channel dimensions: %s", data_shape)

                    if array_key in self.dataset_dtypes:
                        dtype = self.dataset_dtypes[array_key]
                    else:
                        dtype = batch.arrays[array_key].data.dtype

                    logger.debug(
                        "create_dataset: %s, %s, %s, %s, offset=%s, resolution=%s",
                        dataset_name,
                        data_shape,
                        self.compression_type,
                        dtype,
                        offset,
                        voxel_size,
                    )

                    dataset = data_file.create_dataset(
                        name=dataset_name,
                        shape=data_shape,
                        compression=self.compression_type,
                        dtype=dtype,
                    )

                    self._set_offset(dataset, offset)
                    self._set_voxel_size(dataset, voxel_size)

                logger.debug(
                    "%s (%s in %s) has offset %s",
                    array_key,
                    dataset_name,
                    self.store,
                    offset,
                )
                self.dataset_offsets[array_key] = offset

    def process(self, batch, request):
        if not self.dataset_offsets:
            self.init_datasets(batch)

        with self._open_file(self.store) as data_file:
            for array_key, dataset_name in self.dataset_names.items():
                dataset = data_file[dataset_name]

                array_roi = batch.arrays[array_key].spec.roi
                voxel_size = self.spec[array_key].voxel_size
                dims = array_roi.dims
                channel_slices = (slice(None),) * max(0, len(dataset.shape) - dims)

                dataset_roi = Roi(
                    self.dataset_offsets[array_key],
                    Coordinate(dataset.shape[-dims:]) * voxel_size,
                )
                common_roi = array_roi.intersect(dataset_roi)

                if common_roi.empty:
                    logger.warn(
                        "array %s with ROI %s lies outside of dataset ROI %s, "
                        "skipping writing" % (array_key, array_roi, dataset_roi)
                    )
                    continue

                dataset_voxel_roi = (
                    common_roi - self.dataset_offsets[array_key]
                ) // voxel_size
                dataset_voxel_slices = dataset_voxel_roi.to_slices()
                array_voxel_roi = (common_roi - array_roi.offset) // voxel_size
                array_voxel_slices = array_voxel_roi.to_slices()

                logger.debug(
                    "writing %s to voxel coordinates %s"
                    % (array_key, dataset_voxel_roi)
                )

                data = batch.arrays[array_key].data[channel_slices + array_voxel_slices]
                dataset[channel_slices + dataset_voxel_slices] = data



import logging

import numpy as np

from gunpowder.array import ArrayKey
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.roi import Roi

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class Pad(BatchFilter):
    """Add a constant intensity padding around arrays of another batch
    provider. This is useful if your requested batches can be larger than what
    your source provides.

    Args:

        key (:class:`ArrayKey` or :class:`GraphKey`):

            The array or points set to pad.

        size (:class:`Coordinate` or ``None``):

            The padding to be added. If None, an infinite padding is added. If
            a coordinate, this amount will be added to the ROI in the positive
            and negative direction.

        mode (string):

            One of 'constant' or 'reflect'.
            Default is 'constant'

        value (scalar or ``None``):

            The value to report inside the padding. If not given, 0 is used.
            Only used in case of 'constant' mode.
            Only used for :class:`Array<Arrays>`.
    """

    def __init__(self, key, size, mode="constant", value=None):
        self.key = key
        self.size = size
        self.mode = mode
        if self.mode not in ["constant", "reflect"]:
            raise ValueError(
                "Invalid padding mode %s provided. Must be 'constant' or 'reflect'."
                % self.mode
            )
        self.value = value

    def setup(self):
        self.enable_autoskip()

        assert self.key in self.spec, (
            "Asked to pad %s, but is not provided upstream." % self.key
        )
        assert self.spec[self.key].roi is not None, (
            "Asked to pad %s, but upstream provider doesn't have a ROI for "
            "it." % self.key
        )

        spec = self.spec[self.key].copy()
        if self.size is not None:
            spec.roi = spec.roi.grow(self.size, self.size)
        else:
            spec.roi.shape = Coordinate((None,) * spec.roi.dims)
        self.updates(self.key, spec)

    def prepare(self, request):
        upstream_spec = self.get_upstream_provider().spec

        logger.debug("request: %s" % request)
        logger.debug("upstream spec: %s" % upstream_spec)

        # TODO: remove this?
        if self.key not in request:
            return

        roi = request[self.key].roi.copy()

        # change request to fit into upstream spec
        request[self.key].roi = roi.intersect(upstream_spec[self.key].roi)

        if request[self.key].roi.empty:
            logger.warning(
                "Requested %s ROI %s lies entirely outside of upstream ROI %s.",
                self.key,
                roi,
                upstream_spec[self.key].roi,
            )

            # ensure a valid request by asking for empty ROI
            request[self.key].roi = Roi(
                upstream_spec[self.key].roi.offset,
                (0,) * upstream_spec[self.key].roi.dims,
            )

        logger.debug("new request: %s" % request)

        deps = BatchRequest()
        deps[self.key] = request[self.key]
        return deps

    def process(self, batch, request):
        if self.key not in request:
            return

        # restore requested batch size and ROI
        if isinstance(self.key, ArrayKey):
            array = batch.arrays[self.key]
            array.data = self.__expand(
                array.data,
                array.spec.roi / array.spec.voxel_size,
                request[self.key].roi / array.spec.voxel_size,
                self.value if self.value else 0,
            )
            array.spec.roi = request[self.key].roi

        else:
            points = batch.graphs[self.key]
            points.spec.roi = request[self.key].roi

    def __expand(self, a, from_roi, to_roi, value):
        """from_roi and to_roi should be in voxels."""

        logger.debug(
            "expanding array of shape %s from %s to %s", str(a.shape), from_roi, to_roi
        )

        num_channels = len(a.shape) - from_roi.dims
        lower_pad = from_roi.begin - to_roi.begin
        upper_pad = to_roi.end - from_roi.end
        pad_width = [(0, 0)] * num_channels + list(zip(lower_pad, upper_pad))
        if self.mode == "reflect":
            padded = np.pad(a, pad_width, "reflect")
        elif self.mode == "constant":
            padded = np.pad(a, pad_width, "constant", constant_values=value)
        else:
            raise ValueError(
                "Invalid padding mode %s provided. Must be 'constant' or 'reflect'."
                % self.mode
            )
        return padded



import logging
import math
import random
from typing import Optional, Sequence

import numpy as np
from augment.augment import apply_transformation, upscale_transformation
from augment.transform import (
    create_3D_rotation_transformation,
    create_elastic_transformation,
    create_identity_transformation,
    create_rotation_transformation,
)
from scipy import ndimage
from scipy.spatial.transform import Rotation

from gunpowder.array import Array, ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.batch import Batch
from gunpowder.batch_request import BatchRequest
from gunpowder.coordinate import Coordinate
from gunpowder.roi import Roi

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)

# TODO: Add half voxel to points


class DeformAugment(BatchFilter):
    """Elasticly deform a batch. Requests larger batches upstream to avoid data
    loss due to rotation and jitter.

    Args:

        control_point_spacing (``tuple`` of ``int``):

            Distance between control points for the elastic deformation, in
            physical units per dimension.

        jitter_sigma (``Sequence`` of ``float``):

            Standard deviation of control point jitter distribution, in physical units
            per dimension.

        scale_interval (``tuple`` of two ``floats``):

            Interval to randomly sample scale factors from.

        subsample (``int``):

            Instead of creating an elastic transformation on the full
            resolution, create one subsampled by the given factor, and linearly
            interpolate to obtain the full resolution transformation. This can
            significantly speed up this node, at the expense of having visible
            piecewise linear deformations for large factors. Usually, a factor
            of 4 can savely by used without noticable changes. However, the
            default is 1 (i.e., no subsampling).

        spatial_dims (``int``):

            The number of spatial dimensions in arrays. Spatial dimensions are
            assumed to be the last ones and cannot be more than 3 (default).
            Set this value here to avoid treating channels as spacial
            dimension. If, for example, your array is indexed as ``(c,y,x)``
            (2D plus channels), you would want to set ``spatial_dims=2`` to
            perform the elastic deformation only on x and y.

        use_fast_points_transform (``bool``):

            By solving for all of your points simultaneously with the following
            3 step proceedure:
            1) Rasterize nodes into numpy array
            2) Apply elastic transform to array
            3) Read out nodes via center of mass of transformed points
            You can gain substantial speed up as opposed to calculating the
            elastic transform for each point individually. However this may
            lead to nodes being lost during the transform.

        recompute_missing_points (``bool``):

            Whether or not to compute the elastic transform node wise for nodes
            that were lossed during the fast elastic transform process.


        p (``float``, optional):

            Probability applying the augmentation. Default is 1.0 (always
            apply). Should be a float value between 0 and 1. Lowering this value
            could be useful for computational efficiency and increasing
            augmentation space.
    """

    def __init__(
        self,
        control_point_spacing: Coordinate,
        jitter_sigma: Sequence[float],
        scale_interval=(1.0, 1.0),
        rotate: bool = True,
        subsample=1,
        spatial_dims=3,
        use_fast_points_transform=False,
        recompute_missing_points=True,
        transform_key: Optional[ArrayKey] = None,
        graph_raster_voxel_size: Optional[Coordinate] = None,
        p: float = 1.0,
        rotation_axes: Optional[Sequence[int]] = None,
    ):
        self.control_point_spacing = Coordinate(control_point_spacing)
        self.jitter_sigma = np.array(jitter_sigma)
        self.scale_min = scale_interval[0]
        self.scale_max = scale_interval[1]
        self.rotate = rotate
        self.subsample = subsample
        self.spatial_dims = spatial_dims
        self.use_fast_points_transform = use_fast_points_transform
        self.recompute_missing_points = recompute_missing_points
        self.transform_key = transform_key
        self.graph_raster_voxel_size = (
            Coordinate(graph_raster_voxel_size)
            if graph_raster_voxel_size is not None
            else control_point_spacing * 0 + 1
        )
        self.rotation_axes = (
            list(sorted(rotation_axes)) if rotation_axes is not None else None
        )
        assert (
            self.control_point_spacing.dims
            == len(self.jitter_sigma)
            == self.graph_raster_voxel_size.dims
        ), (
            f"control_point_spacing: {self.control_point_spacing}, "
            f"jitter_sigma: {self.jitter_sigma}, "
            f"and graph_raster_voxel_size must have the same number of dimensions"
        )
        if rotation_axes is not None:
            assert (
                2 <= len(rotation_axes) <= spatial_dims <= 3
            ), "Can only rotate in spatial dimensions and only in 2 or 3 dimensions"
            assert all(a < spatial_dims for a in rotation_axes), (
                "axis indexing must be relative to spatial dims. i.e. given b,c,z,y,x "
                "rotation_axes must be a subset of [0,1,2] referencing z,y,x axes"
            )
        self.p = p

    def setup(self):
        if self.transform_key is not None:
            upstream_roi = self.spec.get_total_roi()
            upstream_roi = Roi(
                upstream_roi.offset[-self.spatial_dims :],
                upstream_roi.shape[-self.spatial_dims :],
            ).snap_to_grid(self.control_point_spacing, mode="shrink")
            spec = ArraySpec(
                roi=upstream_roi,
                voxel_size=self.control_point_spacing,
                interpolatable=True,
            )

            self.provides(self.transform_key, spec)

    def skip_node(self, request):
        return random.random() > self.p

    def prepare(self, request):
        # get the total ROI of all requests
        total_roi = request.get_total_roi()
        logger.debug("total ROI is %s" % total_roi)

        # First, get the total ROI of the request in spatial dimensions only.
        # Channels and time don't matter. This is our master ROI.

        # get master ROI
        master_roi = Roi(
            total_roi.begin[-self.spatial_dims :],
            total_roi.shape[-self.spatial_dims :],
        )
        self.spatial_dims = master_roi.dims
        logger.debug("master ROI is %s" % master_roi)

        # make sure the master ROI aligns with the control point spacing
        master_roi_snapped = master_roi.snap_to_grid(
            self.control_point_spacing, mode="grow"
        )
        logger.debug(
            "master ROI aligned with control points is %s" % master_roi_snapped
        )

        # grow by 1 control point spacing
        master_roi_snapped = master_roi_snapped.grow(
            self.control_point_spacing, self.control_point_spacing
        )

        # get master roi in control point spacing
        master_roi_sampled = master_roi_snapped / self.control_point_spacing
        logger.debug("master ROI in control point spacing is %s" % master_roi_sampled)

        # Second, create a master transformation. This is a transformation that
        # covers all voxels of the all requested ROIs. The master transformation
        # is zero-based, all transformations are relative to the origin of master_roi_snapped
        self.master_transformation_spec = ArraySpec(
            master_roi_snapped, self.control_point_spacing, interpolatable=True
        )
        (
            self.master_transformation,
            self.local_transformation,
        ) = self.__create_transformation(self.master_transformation_spec)

        # Third, sample the master transformation for each of the
        # smaller requested ROIs at their respective voxel resolution.
        # crop the parts corresponding to the requested ROIs
        self.transformations = {}
        deps = BatchRequest()
        for key, spec in request.items():
            if key == self.transform_key:
                continue
            spec = spec.copy()

            if spec.roi is None:
                continue

            # get target roi and target spacing (voxel size for arrays or just control point
            # spacing for graphs)
            target_roi = Roi(
                spec.roi.begin[-self.spatial_dims :],
                spec.roi.shape[-self.spatial_dims :],
            )

            # get voxel size of arrays or use graph_raster_voxel_size for graphs
            if isinstance(key, ArrayKey):
                voxel_size = Coordinate(self.spec[key].voxel_size)
            else:
                # must select voxel size for the graph spec because otherwise we would
                # interpolate the transformation onto a spacing of 1 which may be
                # way too large
                voxel_size = self.graph_raster_voxel_size

                # grow target_roi by 1 voxel, this allows us catch nodes that project
                # outside our bounds
                target_roi = target_roi.grow(voxel_size, voxel_size)
                assert (
                    voxel_size is not None
                ), "Please provide a graph_raster_voxel_size when deforming graphs"

            # use only spatial dims for transformations
            voxel_size = Coordinate(voxel_size[-self.spatial_dims :])
            target_spatial_roi = Roi(
                target_roi.offset[-self.spatial_dims :],
                target_roi.shape[-self.spatial_dims :],
            )
            transform_spec = ArraySpec(
                target_spatial_roi.snap_to_grid(voxel_size), voxel_size
            )

            # we save transformations that have been sampled for specific ROI's and voxel sizes,
            # no need to recompute. This can save time if you are requesting multiple arrays of
            # the same voxel size and shape
            if (
                target_spatial_roi.offset,
                target_spatial_roi.shape,
                voxel_size,
            ) in self.transformations:
                transformation = self.transformations[
                    (target_spatial_roi.offset, target_spatial_roi.shape, voxel_size)
                ]
            else:
                # sample the master transformation at the voxel spacing of each array
                transformation = self.__sample_transform(
                    self.master_transformation, transform_spec
                )
                self.transformations[
                    (target_spatial_roi.offset, target_spatial_roi.shape, voxel_size)
                ] = transformation

            # get ROI of all control points necessary to perform transformation
            #
            # for that we follow the same transformations to get from the
            # request ROI to the target ROI in master ROI in control points, just in
            # reverse
            source_roi = self.__get_source_roi(transformation)

            # update upstream request
            spec.roi = Roi(
                spec.roi.begin[: -self.spatial_dims]
                + source_roi.begin[-self.spatial_dims :],
                spec.roi.shape[: -self.spatial_dims]
                + source_roi.shape[-self.spatial_dims :],
            )

            deps[key] = spec

            logger.debug("upstream request roi for %s = %s" % (key, spec.roi))

        return deps

    def process(self, batch, request):
        out_batch = Batch()
        for array_key, array in batch.arrays.items():
            request_roi = Roi(
                request[array_key].roi.offset[-self.spatial_dims :],
                request[array_key].roi.shape[-self.spatial_dims :],
            )
            voxel_size = Coordinate(array.spec.voxel_size[-self.spatial_dims :])
            assert (
                request_roi.offset,
                request_roi.shape,
                voxel_size,
            ) in self.transformations, f"{(request_roi.offset, request_roi.shape, voxel_size)} not in {list(self.transformations.keys())}"

            # reshape array data into (channels,) + spatial dims
            transformed_array = self.__apply_transform(
                array,
                self.transformations[
                    (request_roi.offset, request_roi.shape, voxel_size)
                ],
            )

            out_batch[array_key] = transformed_array

        for graph_key, graph in batch.graphs.items():
            target_spatial_roi = Roi(
                request[graph_key].roi.offset[-self.spatial_dims :],
                request[graph_key].roi.shape[-self.spatial_dims :],
            )
            transform_roi = target_spatial_roi.grow(
                self.graph_raster_voxel_size, self.graph_raster_voxel_size
            )
            source_roi = Roi(
                graph.spec.roi.offset[-self.spatial_dims :],
                graph.spec.roi.shape[-self.spatial_dims :],
            )
            nodes = list(graph.nodes)

            if self.use_fast_points_transform:
                missed_nodes = self.__fast_point_projection(
                    self.transformations[
                        transform_roi.offset,
                        transform_roi.shape,
                        self.graph_raster_voxel_size,
                    ],
                    nodes,
                    source_roi,
                    target_roi=transform_roi,
                )
                if not self.recompute_missing_points:
                    for node in set(missed_nodes):
                        graph.remove_node(node, retain_connectivity=True)
                    missed_nodes = []
            else:
                missed_nodes = nodes

            for node in missed_nodes:
                # logger.debug("projecting %s", node.location)

                # get location relative to beginning of upstream ROI
                location = node.location

                # get spatial coordinates of node
                location_spatial = location[-self.spatial_dims :]

                # get projected location in transformation data space, this
                # yields voxel coordinates relative to target ROI
                projected = self.__project(
                    self.transformations[
                        transform_roi.offset,
                        transform_roi.shape,
                        self.graph_raster_voxel_size,
                    ],
                    location_spatial,
                )

                logger.debug("projected: %s", projected)

                # update spatial coordinates of node location
                node.location[-self.spatial_dims :] = projected

                logger.debug("final location: %s", node.location)

            target_roi = request[graph_key].roi
            logger.debug("Cropping graph to target roi %s", target_roi)
            out_batch[graph_key] = graph.crop(target_roi)

        if self.transform_key is not None:
            out_batch[self.transform_key] = self.local_transformation

        return out_batch

    def __apply_transform(self, array: Array, transformation: Array) -> Array:
        input_shape = array.data.shape
        output_shape = transformation.data.shape
        channel_shape = input_shape[: -self.spatial_dims]
        data = array.data.reshape((-1,) + input_shape[-self.spatial_dims :])

        offset = array.spec.roi.offset[-self.spatial_dims :]
        voxel_size = array.spec.voxel_size[-self.spatial_dims :]

        # apply transformation on each channel
        transform = transformation.data.copy()
        transform -= np.array(offset).reshape((-1,) + (1,) * self.spatial_dims)
        transform /= np.array(voxel_size).reshape((-1,) + (1,) * self.spatial_dims)

        data = np.array(
            [
                apply_transformation(
                    data[c],
                    transform,
                    interpolate=array.spec.interpolatable,
                )
                for c in range(data.shape[0])
            ]
        )
        spec = array.spec.copy()
        spec.roi = Roi(
            spec.roi.offset[: -self.spatial_dims] + transformation.spec.roi.offset[:],
            spec.roi.shape[: -self.spatial_dims] + transformation.spec.roi.shape[:],
        )

        return Array(
            data.reshape(channel_shape + output_shape[-self.spatial_dims :]), spec
        )

    def __sample_transform(
        self,
        transformation: Array,
        output_spec: ArraySpec,
        interpolate_order=1,
    ) -> Array:
        if output_spec.voxel_size == transformation.spec.voxel_size:
            # if voxel_size == control_point_spacing we can simply slice into the master roi
            relative_output_roi = (
                output_spec.roi - transformation.spec.roi.offset
            ).snap_to_grid(output_spec.voxel_size) / output_spec.voxel_size
            sampled = np.copy(
                transformation.data[
                    (slice(None),) + relative_output_roi.get_bounding_box()
                ]
            )
            return Array(
                sampled,
                ArraySpec(
                    output_spec.roi.snap_to_grid(output_spec.voxel_size),
                    output_spec.voxel_size,
                    interpolatable=True,
                ),
            )

        dims = len(output_spec.voxel_size)
        output_shape = output_spec.roi.shape / output_spec.voxel_size
        offset = np.array(
            [
                o / s
                for o, s in zip(
                    output_spec.roi.offset - transformation.spec.roi.offset,
                    transformation.spec.voxel_size,
                )
            ]
        )
        step = np.array(
            [
                o / i
                for o, i in zip(output_spec.voxel_size, transformation.spec.voxel_size)
            ]
        )
        coordinates = np.stack(
            np.meshgrid(
                range(dims),
                *[
                    np.linspace(o, (shape - 1) * step + o, shape)
                    for o, shape, step in zip(offset, output_shape, step)
                ],
                indexing="ij",
            )
        )

        sampled = ndimage.map_coordinates(
            transformation.data,
            coordinates=coordinates,
            order=3,
            mode="nearest",
        )
        return Array(sampled, ArraySpec(output_spec.roi, output_spec.voxel_size))

    def __create_transformation(self, target_spec: ArraySpec):
        scale = self.scale_min + random.random() * (self.scale_max - self.scale_min)

        target_shape = target_spec.roi.shape / target_spec.voxel_size

        global_transformation = create_identity_transformation(
            target_shape,
            subsample=self.subsample,
            scale=scale,
        )
        local_transformation = np.zeros_like(global_transformation)

        if sum(self.jitter_sigma) > 0:
            el_transformation = create_elastic_transformation(
                target_shape,
                np.array(self.control_point_spacing) / np.array(target_spec.voxel_size),
                np.array(self.jitter_sigma) / np.array(target_spec.voxel_size),
                subsample=self.subsample,
            )

            local_transformation += el_transformation

        if self.rotate:
            assert min(target_spec.voxel_size) == max(
                target_spec.voxel_size
            ), "Only isotropic control point spacing supported when rotating"

            if self.spatial_dims == 2 or (
                self.rotation_axes is not None and len(self.rotation_axes) == 2
            ):
                if self.rotation_axes is not None:
                    target_rot_shape = tuple(np.array(target_shape)[self.rotation_axes])
                else:
                    target_rot_shape = target_shape
                rot_transformation = create_rotation_transformation(
                    target_rot_shape,
                    random.random() * math.pi * 2,
                    subsample=self.subsample,
                )
                if self.rotation_axes:
                    slices = (slice(None),) + tuple(
                        slice(None) if i in self.rotation_axes else np.newaxis
                        for i in range(self.spatial_dims)
                    )
                    rot_transformation = rot_transformation[slices]
                    # figure out which axes were skipped. i.e. [0,-1,2] means axes 0,2 were rotated
                    dzyx_sampling = [
                        self.rotation_axes.index(j) if j in self.rotation_axes else -1
                        for j in range(self.spatial_dims)
                    ]
                    # stack the rotation transformation to have the appropriate number of offset dims
                    # inserting 0s for the skipped axis
                    rot_transformation = np.stack(
                        [
                            (
                                rot_transformation[x]
                                if x >= 0
                                else np.zeros_like(rot_transformation[0])
                            )
                            for x in dzyx_sampling
                        ],
                        axis=0,
                    )

            else:
                angle = Rotation.random()
                rot_transformation = create_3D_rotation_transformation(
                    target_shape, angle, subsample=self.subsample
                )

            local_transformation += rot_transformation

        if self.subsample > 1:
            global_transformation = upscale_transformation(
                global_transformation, target_shape
            )
            local_transformation = upscale_transformation(
                local_transformation, target_shape
            )
            global_transformation = upscale_transformation(
                global_transformation, target_shape
            )

        # transform into world units
        global_transformation *= np.array(target_spec.voxel_size).reshape(
            (len(target_spec.voxel_size),) + (1,) * self.spatial_dims
        )
        global_transformation += np.array(target_spec.roi.offset).reshape(
            (len(target_spec.roi.offset),) + (1,) * self.spatial_dims
        )

        local_transformation *= np.array(target_spec.voxel_size).reshape(
            (len(target_spec.voxel_size),) + (1,) * self.spatial_dims
        )

        return (
            Array(global_transformation + local_transformation, target_spec),
            Array(local_transformation, target_spec),
        )

    def __fast_point_projection(self, transformation, nodes, source_roi, target_roi):
        valid_nodes = [node for node in nodes if source_roi.contains(node.location)]
        if len(valid_nodes) < 1:
            return []
        # rasterize the points into an array
        ids, locs = zip(
            *[
                (
                    node.id,
                    (
                        np.floor(node.location[-self.spatial_dims :]).astype(int)
                        - source_roi.begin
                    )
                    // self.graph_raster_voxel_size,
                )
                for node in valid_nodes
            ]
        )
        ids, locs = np.array(ids), tuple(zip(*locs))
        points_array = np.zeros(
            source_roi.shape / self.graph_raster_voxel_size, dtype=np.int64
        )
        points_array[locs] = ids

        # reshape array data into (channels,) + spatial dims
        shape = points_array.shape
        data = points_array.reshape((-1,) + shape[-self.spatial_dims :])

        array = Array(
            data,
            ArraySpec(
                Roi(
                    source_roi.begin[-self.spatial_dims :],
                    Coordinate(shape) * self.graph_raster_voxel_size,
                ),
                self.graph_raster_voxel_size,
            ),
        )
        transformed = self.__apply_transform(array, transformation)

        data = transformed.data
        missing_points = []
        projected_locs = ndimage.center_of_mass(data > 0, data, ids)
        projected_locs = [
            (np.array(loc[-self.spatial_dims :]) + 0.5) * self.graph_raster_voxel_size
            + transformation.spec.roi.begin
            for loc in projected_locs
        ]
        node_dict = {node.id: node for node in nodes}
        for point_id, proj_loc in zip(ids, projected_locs):
            point = node_dict.pop(point_id)
            if not any([np.isnan(x) for x in proj_loc]):
                assert (
                    len(proj_loc) == self.spatial_dims
                ), "projected location has wrong number of dimensions: {}, expected: {}".format(
                    len(proj_loc), self.spatial_dims
                )
                point.location[-self.spatial_dims :] = proj_loc
            else:
                missing_points.append(point)
        for node in node_dict.values():
            missing_points.append(node)
        logger.debug(
            "{} of {} points lost in fast points projection".format(
                len(missing_points), len(ids)
            )
        )

        return missing_points

    def __project(self, transformation: Array, location: np.ndarray) -> np.ndarray:
        """Find the projection of location given by transformation. Returns None
        if projection lies outside of transformation."""

        dims = len(location)

        # subtract location from transformation
        diff = transformation.data.copy()
        for d in range(dims):
            diff[d] -= location[d]

        # square
        diff2 = diff * diff

        # sum
        dist = diff2.sum(axis=0)

        # find grid point closes to location
        center_grid = Coordinate(np.unravel_index(dist.argmin(), dist.shape))
        center_source = self.__source_at(transformation, center_grid)

        logger.debug("projecting %s onto grid", location)
        logger.debug("grid shape: %s", transformation.data.shape[1:])
        logger.debug("grid projection: %s", center_grid)
        logger.debug("dist shape: %s", dist.shape)
        logger.debug("dist.argmin(): %s", dist.argmin())
        logger.debug("dist[argmin]: %s", dist[center_grid])
        logger.debug(
            "transform[argmin]: %s", transformation.data[(slice(None),) + center_grid]
        )
        logger.debug("min dist: %s", dist.min())
        logger.debug("center source: %s", center_source)

        # add a half voxel step to localize each transformed point to the center of the
        # closest voxel
        return (
            np.array(center_grid, dtype=np.float32) + 0.5
        ) * transformation.spec.voxel_size + transformation.spec.roi.offset

    def __source_at(self, transformation, index):
        """Read the source point of a transformation at index."""

        slices = (slice(None),) + tuple(slice(i, i + 1) for i in index)
        return transformation.data[slices].flatten()

    def __get_source_roi(self, transformation):
        # this gets you the source_roi in offset space. We need to add 1 voxel
        # to the shape to get the closed interval ROI

        # get bounding box of needed data for transformation
        bb_min = Coordinate(
            int(math.floor(transformation.data[d].min()))
            for d in range(transformation.spec.voxel_size.dims)
        )
        bb_max = Coordinate(
            int(math.ceil(transformation.data[d].max())) + s
            for d, s in zip(
                range(transformation.spec.voxel_size.dims),
                transformation.spec.voxel_size,
            )
        )

        # create roi sufficiently large to feed transformation
        source_roi = Roi(bb_min, bb_max - bb_min).snap_to_grid(
            transformation.spec.voxel_size
        )

        return source_roi

    def __shift_transformation(self, shift, transformation):
        for d in range(transformation.shape[0]):
            transformation[d] += shift[d]



from funlib.persistence.arrays import Array as PersistenceArray

from gunpowder.array import Array, ArrayKey
from gunpowder.array_spec import ArraySpec
from gunpowder.batch import Batch

from .batch_provider import BatchProvider


class ArraySource(BatchProvider):
    """A `array <https://github.com/funkelab/funlib.persistence>`_ source.

    Provides a source for any array that can fit into the funkelab
    funlib.persistence.Array format. This class comes with assumptions about
    the available metadata and convenient methods for indexing the data
    with a :class:`Roi` in world units.

    Args:

        key (:class:`ArrayKey`):

            The ArrayKey for accessing this array.

        array (``Array``):

            A `funlib.persistence.Array` object.

        interpolatable (``bool``, optional):

            Whether the array is interpolatable. If not given it is
            guessed based on dtype.

    """

    def __init__(
        self,
        key: ArrayKey,
        array: PersistenceArray,
        interpolatable: bool | None = None,
    ):
        self.key = key
        self.array = array
        self.array_spec = ArraySpec(
            self.array.roi,
            self.array.voxel_size,
            interpolatable,
            False,
            self.array.dtype,
        )

    def setup(self):
        self.provides(self.key, self.array_spec)

    def provide(self, request):
        outputs = Batch()
        out_spec = self.array_spec.copy()
        out_spec.roi = request[self.key].roi
        outputs[self.key] = Array(self.array[out_spec.roi], out_spec)
        return outputs



import logging
import multiprocessing
from abc import ABC

import numpy as np
import tqdm

from gunpowder.array import Array
from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.graph import Graph
from gunpowder.producer_pool import ProducerPool
from gunpowder.roi import Roi

from .batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class ScanCallback(ABC):
    """Base class for :class:`Scan` callbacks. Implement any of ``start``,
    ``update``, and ``stop`` in a subclass to create your own callback.
    """

    def start(self, num_total):
        """Called once before :class:`Scan` starts scanning over chunks.

        Args:

            num_total (int):

                The total number of chunks to process.
        """
        pass

    def update(self, num_processed):
        """Called periodically by :class:`Scan` while processing chunks.

        Args:

            num_processed (int):

                The number of chunks already processed.
        """
        pass

    def stop(self):
        """Called once after :class:`Scan` scanned over all chunks."""
        pass


class TqdmCallback(ScanCallback):
    """A default callback that uses ``tqdm`` to show a progress bar."""

    def start(self, num_total):
        logger.info("scanning over %d chunks", num_total)

        self.progress_bar = tqdm.tqdm(desc="Scan, chunks processed", total=num_total)
        self.num_processed = 0

    def update(self, num_processed):
        self.progress_bar.update(num_processed - self.num_processed)
        self.num_processed = num_processed

    def stop(self):
        self.progress_bar.close()


class Scan(BatchFilter):
    """Iteratively requests batches of size ``reference`` from upstream
    providers in a scanning fashion, until all requested ROIs are covered. If
    the batch request to this node is empty, it will scan the complete upstream
    ROIs (and return nothing). Otherwise, it scans only the requested ROIs and
    returns a batch assembled of the smaller requests. In either case, the
    upstream requests will be contained in the downstream requested ROI or
    upstream ROIs.

    See also :class:`Hdf5Write`.

    Args:

        reference (:class:`BatchRequest`):

            A reference :class:`BatchRequest`. This request will be shifted in
            a scanning fashion over the upstream ROIs of the requested arrays
            or points.

        num_workers (``int``, optional):

            If set to >1, upstream requests are made in parallel with that
            number of workers.

        cache_size (``int``, optional):

            If multiple workers are used, how many batches to hold at most.

        progress_callback (class:`ScanCallback`, optional):

            A callback instance to get updated from this node while processing
            chunks. See :class:`ScanCallback` for details. The default is a
            callback that shows a ``tqdm`` progress bar.
    """

    def __init__(self, reference, num_workers=1, cache_size=50, progress_callback=None):
        self.reference = reference.copy()
        self.num_workers = num_workers
        self.cache_size = cache_size
        self.workers = None
        self.batch = None
        if progress_callback is None:
            self.progress_callback = TqdmCallback()
        else:
            self.progress_callback = progress_callback

    def setup(self):
        if self.num_workers > 1:
            self.request_queue = multiprocessing.Queue(maxsize=0)
            self.workers = ProducerPool(
                [self._worker_get_chunk for _ in range(self.num_workers)],
                queue_size=self.cache_size,
            )
            self.workers.start()

    def teardown(self):
        if self.num_workers > 1:
            self.workers.stop()

    def provide(self, request):
        empty_request = len(request) == 0
        if empty_request:
            scan_spec = self.spec
        else:
            scan_spec = request

        stride = self._get_stride()
        shift_roi = self._get_shift_roi(scan_spec)

        shifts = self._enumerate_shifts(shift_roi, stride)
        num_chunks = len(shifts)

        if self.progress_callback is not None:
            self.progress_callback.start(num_chunks)

        # the batch to return
        self.batch = Batch()

        if self.num_workers > 1:
            for shift in shifts:
                shifted_reference = self._shift_request(self.reference, shift)
                self.request_queue.put(shifted_reference)

            for i in range(num_chunks):
                chunk = self.workers.get()

                if not empty_request:
                    self._add_to_batch(request, chunk)

                if self.progress_callback is not None:
                    self.progress_callback.update(i + 1)

                logger.debug("processed chunk %d/%d", i + 1, num_chunks)

        else:
            for i, shift in enumerate(shifts):
                shifted_reference = self._shift_request(self.reference, shift)
                chunk = self._get_chunk(shifted_reference)

                if not empty_request:
                    self._add_to_batch(request, chunk)

                if self.progress_callback is not None:
                    self.progress_callback.update(i + 1)

                logger.debug("processed chunk %d/%d", i + 1, num_chunks)

        if self.progress_callback is not None:
            self.progress_callback.stop()

        batch = self.batch
        self.batch = None

        logger.debug("returning batch %s", batch)

        return batch

    def _get_stride(self):
        """Get the maximal amount by which ``reference`` can be moved, such
        that it tiles the space."""

        stride = None

        # get the least common multiple of all voxel sizes, we have to stride
        # at least that far
        lcm_voxel_size = self.spec.get_lcm_voxel_size(self.reference.array_specs.keys())

        # that's just the minimal size in each dimension
        for key, reference_spec in self.reference.items():
            shape = reference_spec.roi.shape

            for d in range(len(lcm_voxel_size)):
                assert shape[d] >= lcm_voxel_size[d], (
                    "Shape of reference "
                    "ROI %s for %s is "
                    "smaller than least "
                    "common multiple of "
                    "voxel size "
                    "%s" % (reference_spec.roi, key, lcm_voxel_size)
                )

            if stride is None:
                stride = shape
            else:
                stride = Coordinate((min(a, b) for a, b in zip(stride, shape)))

        return stride

    def _get_shift_roi(self, spec):
        """Get the minimal and maximal shift (as a ROI) to apply to
        ``self.reference``, such that it is still fully contained in ``spec``.
        """

        total_shift_roi = None

        # get individual shift ROIs and intersect them
        for key, reference_spec in self.reference.items():
            logger.debug("getting shift roi for %s with spec %s", key, reference_spec)

            if key not in spec:
                logger.debug("skipping, %s not in upstream spec", key)
                continue
            if spec[key].roi is None:
                logger.debug("skipping, %s has not ROI", key)
                continue

            logger.debug("upstream ROI is %s", spec[key].roi)

            for r, s in zip(reference_spec.roi.shape, spec[key].roi.shape):
                assert s is None or r <= s, (
                    "reference %s with ROI %s does not fit into provided "
                    "upstream %s" % (key, reference_spec.roi, spec[key].roi)
                )

            # we have a reference ROI
            #
            #    [--------) [9]
            #    3        12
            #
            # and a spec ROI
            #
            #                 [---------------) [16]
            #                 16              32
            #
            # min and max shifts of reference are
            #
            #                 [--------) [9]
            #                 16       25
            #                        [--------) [9]
            #                        23       32
            #
            # therefore, all possible ways to shift the reference such that it
            # is contained in the spec are at least 16-3=13 and at most 23-3=20
            # (inclusive)
            #
            #              [-------) [8]
            #              13      21
            #
            # 1. the starting point is beginning of spec - beginning of reference
            # 2. the length is length of spec - length of reference + 1

            # 1. get the starting point of the shift ROI
            shift_begin = spec[key].roi.begin - reference_spec.roi.begin

            # 2. get the shape of the shift ROI
            shift_shape = spec[key].roi.shape - reference_spec.roi.shape + 1

            # create a ROI...
            shift_roi = Roi(shift_begin, shift_shape)

            logger.debug("shift ROI for %s is %s", key, shift_roi)

            # ...and intersect it with previous shift ROIs
            if total_shift_roi is None:
                total_shift_roi = shift_roi
            else:
                total_shift_roi = total_shift_roi.intersect(shift_roi)
                if total_shift_roi.empty:
                    raise RuntimeError(
                        "There is no location where the ROIs "
                        "the reference %s are contained in the "
                        "request/upstream ROIs "
                        "%s." % (self.reference, spec)
                    )

            logger.debug(
                "intersected with total shift ROI this yields %s", total_shift_roi
            )

        if total_shift_roi is None:
            raise RuntimeError(
                "None of the upstream ROIs are bounded (all "
                "ROIs are None). Scan needs at least one "
                "bounded upstream ROI."
            )

        return total_shift_roi

    def _enumerate_shifts(self, shift_roi, stride):
        """Produces a sequence of shift coordinates starting at the beginning
        of ``shift_roi``, progressing with ``stride``. The maximum shift
        coordinate in any dimension will be the last point inside the shift roi
        in this dimension."""

        min_shift = shift_roi.offset
        max_shift = max(min_shift, Coordinate(m - 1 for m in shift_roi.end))

        shift = np.array(min_shift)
        shifts = []

        dims = len(min_shift)

        logger.debug("enumerating possible shifts of %s in %s", stride, shift_roi)

        while True:
            logger.debug("adding %s", shift)
            shifts.append(Coordinate(shift))

            if (shift == max_shift).all():
                break

            # count up dimensions
            for d in range(dims):
                if shift[d] >= max_shift[d]:
                    if d == dims - 1:
                        break
                    shift[d] = min_shift[d]
                else:
                    shift[d] += stride[d]
                    # snap to last possible shift, don't overshoot
                    if shift[d] > max_shift[d]:
                        shift[d] = max_shift[d]
                    break

        return shifts

    def _shift_request(self, request, shift):
        shifted = request.copy()
        for _, spec in shifted.items():
            spec.roi = spec.roi.shift(shift)

        return shifted

    def _worker_get_chunk(self):
        request = self.request_queue.get()
        return self._get_chunk(request)

    def _get_chunk(self, request):
        return self.get_upstream_provider().request_batch(request)

    def _add_to_batch(self, spec, chunk):
        if self.batch.get_total_roi() is None:
            self.batch = self._setup_batch(spec, chunk)
        self.batch.profiling_stats.merge_with(chunk.profiling_stats)

        for array_key, array in chunk.arrays.items():
            if array_key not in spec:
                continue
            self._fill(
                self.batch.arrays[array_key].data,
                array.data,
                spec.array_specs[array_key].roi,
                array.spec.roi,
                self.spec[array_key].voxel_size,
            )

        for graph_key, graphs in chunk.graphs.items():
            if graph_key not in spec:
                continue
            self._fill_points(
                self.batch.graphs[graph_key],
                graphs,
                spec.graph_specs[graph_key].roi,
                graphs.spec.roi,
            )

    def _setup_batch(self, batch_spec, chunk):
        """Allocate a batch matching the sizes of ``batch_spec``, using
        ``chunk`` as template."""

        batch = Batch()

        for array_key, spec in batch_spec.array_specs.items():
            roi = spec.roi
            voxel_size = self.spec[array_key].voxel_size

            # get the 'non-spatial' shape of the chunk-batch
            # and append the shape of the request to it
            array = chunk.arrays[array_key]
            shape = array.data.shape[: -roi.dims]
            shape += roi.shape // voxel_size

            spec = self.spec[array_key].copy()
            spec.roi = roi
            logger.info("allocating array of shape %s for %s", shape, array_key)
            batch.arrays[array_key] = Array(
                data=np.zeros(shape, dtype=spec.dtype), spec=spec
            )

        for graph_key, spec in batch_spec.graph_specs.items():
            roi = spec.roi
            spec = self.spec[graph_key].copy()
            spec.roi = roi
            batch.graphs[graph_key] = Graph(nodes=[], edges=[], spec=spec)

        logger.debug("setup batch to fill %s", batch)

        return batch

    def _fill(self, a, b, roi_a, roi_b, voxel_size):
        logger.debug("filling " + str(roi_b) + " into " + str(roi_a))

        roi_a = roi_a // voxel_size
        roi_b = roi_b // voxel_size

        common_roi = roi_a.intersect(roi_b)
        if common_roi.empty:
            return

        common_in_a_roi = common_roi - roi_a.offset
        common_in_b_roi = common_roi - roi_b.offset

        slices_a = common_in_a_roi.get_bounding_box()
        slices_b = common_in_b_roi.get_bounding_box()

        if len(a.shape) > len(slices_a):
            slices_a = (slice(None),) * (len(a.shape) - len(slices_a)) + slices_a
            slices_b = (slice(None),) * (len(b.shape) - len(slices_b)) + slices_b

        a[slices_a] = b[slices_b]

    def _fill_points(self, a, b, roi_a, roi_b):
        """
        Take points from b and add them to a.
        Nodes marked temporary must be ignored. Temporary nodes are nodes
        that were created during processing. Since it is impossible to know
        in general, that a node created during processing of a subgraph was
        not assigned an id that is already used by the full graph, we cannot
        include temporary nodes and assume there will not be ambiguous node
        id's that correspond to multiple distinct nodes.
        """
        logger.debug("filling points of " + str(roi_b) + " into points of" + str(roi_a))

        common_roi = roi_a.intersect(roi_b)
        if common_roi is None:
            return

        for node in b.nodes:
            if not node.temporary and roi_a.contains(node.location):
                a.add_node(node)
        for e in b.edges:
            bu = b.node(e.u)
            bv = b.node(e.v)
            if (
                not bu.temporary
                and not bv.temporary
                and a.contains(bu.id)
                and a.contains(bv.id)
            ):
                a.add_edge(e)



import copy
import logging

import numpy as np

from gunpowder.array import Array
from gunpowder.batch_request import BatchRequest
from gunpowder.nodes.batch_filter import BatchFilter

logger = logging.getLogger(__name__)


class PrepareMalis(BatchFilter):
    """Creates a component label array needed for two-phase malis training.

    Args:

        labels_array_key(:class:`ArrayKey`): The label array to use.

        malis_comp_array_key(:class:`ArrayKey`): The malis component array
            to generate.

        ignore_array_key(:class:`ArrayKey`, optional): An ignore mask to
            use.
    """

    def __init__(self, labels_array_key, malis_comp_array_key, ignore_array_key=None):
        self.labels_array_key = labels_array_key
        self.malis_comp_array_key = malis_comp_array_key
        self.ignore_array_key = ignore_array_key

    def setup(self):
        spec = self.spec[self.labels_array_key].copy()
        self.provides(self.malis_comp_array_key, spec)
        self.enable_autoskip()

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.labels_array_key] = copy.deepcopy(request[self.labels_array_key])
        if self.ignore_array_key is not None:
            deps[self.ignore_array_key] = copy.deepcopy(request[self.labels_array_key])
        return deps

    def process(self, batch, request):
        gt_labels = batch.arrays[self.labels_array_key]
        next_id = gt_labels.data.max() + 1

        gt_pos_pass = gt_labels.data

        if self.ignore_array_key is not None:
            gt_neg_pass = np.array(gt_labels.data)
            gt_neg_pass[batch.arrays[self.ignore_array_key].data == 0] = next_id

        else:
            gt_neg_pass = gt_pos_pass

        spec = self.spec[self.malis_comp_array_key].copy()
        spec.roi = request[self.labels_array_key].roi
        batch.arrays[self.malis_comp_array_key] = Array(
            np.array([gt_neg_pass, gt_pos_pass]), spec
        )

        # Why don't we update gt_affinities in the same way?
        # -> not needed
        #
        # GT affinities are all 0 in the masked area (because masked area is
        # assumed to be set to background in batch.gt).
        #
        # In the negative pass:
        #
        #   We set all affinities inside GT regions to 1. Affinities in masked
        #   area as predicted. Belongs to one forground region (introduced
        #   above). But we only count loss on edges connecting different labels
        #   -> loss in masked-out area only from outside regions.
        #
        # In the positive pass:
        #
        #   We set all affinities outside GT regions to 0 -> no loss in masked
        #   out area.



import csv
import logging
from typing import Optional, Union

import numpy as np

from gunpowder.batch import Batch
from gunpowder.coordinate import Coordinate
from gunpowder.graph import Graph, GraphKey, Node
from gunpowder.graph_spec import GraphSpec
from gunpowder.nodes.batch_provider import BatchProvider
from gunpowder.profiling import Timing
from gunpowder.roi import Roi

logger = logging.getLogger(__name__)


class CsvPointsSource(BatchProvider):
    """Read a set of points from a comma-separated-values text file. Each line
    in the file represents one point, e.g. z y x (id). Note: this reads all
    points into memory and finds the ones in the given roi by iterating
    over all the points. For large datasets, this may be too slow.

    Args:

        filename (``string``):

            The file to read from.

        points (:class:`GraphKey`):

            The key of the points set to create.

        spatial_cols (list[``int``]):

            The columns of the csv that hold the coordinates of the points
            (in the order that you want them to be used in training)

        points_spec (:class:`GraphSpec`, optional):

            An optional :class:`GraphSpec` to overwrite the points specs
            automatically determined from the CSV file. This is useful to set
            the :class:`Roi` manually.

        scale (scalar or array-like):

            An optional scaling to apply to the coordinates of the points read
            from the CSV file. This is useful if the points refer to voxel
            positions to convert them to world units.

        id_col (``int``, optional):

            The column of the csv that holds an id for each point. If not
            provided, the index of the rows are used as the ids. When read
            from file, ids are left as strings and not cast to anything.

        delimiter (``str``, optional):

            Delimiter to pass to the csv reader. Defaults to ",".
    """

    def __init__(
        self,
        filename: str,
        points: GraphKey,
        spatial_cols: list[int],
        points_spec: Optional[GraphSpec] = None,
        scale: Optional[Union[int, float, tuple, list, np.ndarray]] = None,
        id_col: Optional[int] = None,
        delimiter: str = ",",
    ):
        self.filename = filename
        self.points = points
        self.points_spec = points_spec
        self.scale = scale
        self.spatial_cols = spatial_cols
        self.id_dim = id_col
        self.delimiter = delimiter
        self.data: Optional[np.ndarray] = None
        self.ids: Optional[list] = None

    def setup(self):
        self._parse_csv()

        if self.points_spec is not None:
            self.provides(self.points, self.points_spec)
            return

        min_bb = Coordinate(np.floor(np.amin(self.data, 0)))
        max_bb = Coordinate(np.ceil(np.amax(self.data, 0)) + 1)

        roi = Roi(min_bb, max_bb - min_bb)

        self.provides(self.points, GraphSpec(roi=roi))

    def provide(self, request):
        timing = Timing(self)
        timing.start()

        min_bb = request[self.points].roi.begin
        max_bb = request[self.points].roi.end

        logger.debug("CSV points source got request for %s", request[self.points].roi)

        point_filter = np.ones((self.data.shape[0],), dtype=bool)
        for d in range(len(self.spatial_cols)):
            point_filter = np.logical_and(point_filter, self.data[:, d] >= min_bb[d])
            point_filter = np.logical_and(point_filter, self.data[:, d] < max_bb[d])

        points_data = self._get_points(point_filter)
        points_spec = GraphSpec(roi=request[self.points].roi.copy())

        batch = Batch()
        batch.graphs[self.points] = Graph(points_data, [], points_spec)

        timing.stop()
        batch.profiling_stats.add(timing)

        return batch

    def _get_points(self, point_filter):
        filtered = self.data[point_filter]
        ids = self.ids[point_filter]
        return [Node(id=i, location=p) for i, p in zip(ids, filtered)]

    def _parse_csv(self):
        """Read one point per line, with spatial and id columns determined by
        self.spatial_cols and self.id_col.
        """
        data = []
        ids = []
        with open(self.filename, "r", newline="") as f:
            has_header = csv.Sniffer().has_header(f.read(1024))
            f.seek(0)
            first_line = True
            reader = csv.reader(f, delimiter=self.delimiter)
            for line in reader:
                if first_line and has_header:
                    first_line = False
                    continue
                space = [float(line[c]) for c in self.spatial_cols]
                data.append(space)
                if self.id_dim is not None:
                    ids.append(line[self.id_dim])

        self.data = np.array(data, dtype=np.float32)
        if self.id_dim:
            self.ids = np.array(ids)
        else:
            self.ids = np.arange(len(self.data))

        if self.scale is not None:
            self.data *= self.scale



import itertools
import random
import warnings

import numpy as np

from gunpowder.batch_request import BatchRequest

from .batch_filter import BatchFilter


class IntensityAugment(BatchFilter):
    """Randomly scale and shift the values of an intensity array.

    Args:

        array (:class:`ArrayKey`):

            The intensity array to modify.

        scale_min (``float``):
        scale_max (``float``):
        shift_min (``float``):
        shift_max (``float``):

            The min and max of the uniformly randomly drawn scaling and
            shifting values for the intensity augmentation. Intensities are
            changed as::

                a = a.mean() + (a-a.mean())*scale + shift

        z_section_wise (``bool``):

            Perform the augmentation z-section wise. Requires 3D arrays and
            assumes that z is the first dimension.

        clip (``bool``):

            Set to False if modified values should not be clipped to [0, 1]
            Disables range check!

        p (``float``, optional):

            Probability applying the augmentation. Default is 1.0 (always
            apply). Should be a float value between 0 and 1. Lowering this value
            could be useful for computational efficiency and increasing
            augmentation space.

        slab (``tuple`` of ``int``, optional):

            A shape specification to perform the intensity augment in slabs of this
            size. -1 can be used to refer to the actual size of the label
            array. For example, a slab of::

                (2, -1, -1, -1)

            will perform the intensity augment for every each slice ``[0:2,:]``,
            ``[2:4,:]``, ... individually on 4D data.
    """

    def __init__(
        self,
        array,
        scale_min,
        scale_max,
        shift_min,
        shift_max,
        z_section_wise=False,
        clip=True,
        p=1.0,
        slab=None,
    ):
        self.array = array
        self.scale_min = scale_min
        self.scale_max = scale_max
        self.shift_min = shift_min
        self.shift_max = shift_max
        self.z_section_wise = z_section_wise
        if self.z_section_wise:
            warnings.warn(
                DeprecationWarning(
                    "z_section_wise is deprecated and will be removed in the future. Please use slab instead."
                )
            )
            assert slab is None, "z_section_wise and slab are mutually exclusive."

        self.clip = clip
        self.p = p
        self.slab = slab

    def setup(self):
        self.enable_autoskip()
        self.updates(self.array, self.spec[self.array])

    def skip_node(self, request):
        return random.random() > self.p

    def prepare(self, request):
        deps = BatchRequest()
        deps[self.array] = request[self.array].copy()
        return deps

    def process(self, batch, request):
        raw = batch.arrays[self.array]

        assert raw.data.dtype == np.float32 or raw.data.dtype == np.float64, (
            "Intensity augmentation requires float types for the raw array (not "
            + str(raw.data.dtype)
            + "). Consider using Normalize before."
        )
        if self.clip:
            assert (
                raw.data.min() >= 0 and raw.data.max() <= 1
            ), "Intensity augmentation expects raw values in [0,1]. Consider using Normalize before."

        if self.z_section_wise is not None:
            slab = [-1] * len(raw.data.shape)
            slab[-raw.spec.voxel_size.dims] = 1
        elif self.slab is not None:
            slab = self.slab
        else:
            slab = [-1] * len(raw.data.shape)

        # slab with -1 replaced by shape
        slab = tuple(m if s == -1 else s for m, s in zip(raw.data.shape, slab))

        slab_ranges = (range(0, m, s) for m, s in zip(raw.data.shape, slab))

        for start in itertools.product(*slab_ranges):
            slices = tuple(
                slice(start[d], start[d] + slab[d]) for d in range(len(slab))
            )
            raw.data[slices] = self.__augment(
                raw.data[slices],
                np.random.uniform(low=self.scale_min, high=self.scale_max),
                np.random.uniform(low=self.shift_min, high=self.shift_max),
            )

        # clip values, we might have pushed them out of [0,1]
        if self.clip:
            raw.data[raw.data > 1] = 1
            raw.data[raw.data < 0] = 0

    def __augment(self, a, scale, shift):
        return a.mean() + (a - a.mean()) * scale + shift
